/* **********************************************************
 *   Program:    PC Keyboard
 *   wb - 01.Oct.2006 -- KBD-Send for LEDs
 *   wb - 09.Mar.2005 -- ShiftLock-Behandlung
 *   wb - 29.Jan.2005 -- Based on ...
 *        TODO  - Status-LEDs auf Tastatur
 *
 *   Created:    07.09.99 21:05
 *   Author:     V. Brajer - vlado.brajer@kks.s-net.net
 *               http://www.sparovcek.net/bray
 *               based on Atmel's AVR313 application note
 *   Comments:   AVRGCC port - original for IAR C
 *********************************************************** */
#include <avr/io.h>
#include <avr/interrupt.h>
#if (__AVR_LIBC_VERSION__ < 10400UL)
#include <avr/signal.h>
#endif
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "pckb.h"
#include "scancodes.h"
#include "uart.h"

/* -------------------- Portability macros ------------------------ */
// Keybord Input Data-Pin
#define DDR_KB_DATA         DDRD
#define PIN_KB_DATA         PIND
#define KB_DATA_BITNR       PD4
#define KB_DATA_BIT         (1<<KB_DATA_BITNR)
#define KB_DATA_GET()     (PIN_KB_DATA & KB_DATA_BIT)

// Keyboard Output data-Port
#define DDR_KB_DATA_OUT     DDRD
#define PORT_KB_DATA_OUT    PORTD
#define KB_DATA_OUT_BITNR   PD5
#define KB_DATA_OUT_BIT     (1<<KB_DATA_OUT_BITNR)

// Keyboard Clocking
// Interrupt
#define KB_CLOCKINT         INT0
#define KB_INT_MASKREG      GICR
#define KB_INT_FLAGREG      GIFR
#define KB_INTERRUPT        SIG_INTERRUPT0
#define FALLING_EDGE()      MCUCR = (MCUCR & ~((1<<ISC01)|(1<<ISC00)))|(1<<ISC01)|(0<<ISC00)
#define RISING_EDGE()       MCUCR |= (1<<ISC01)|(1<<ISC00)
#define ANY_EDGE()          MCUCR = (MCUCR & ~((1<<ISC01)|(1<<ISC00)))|(0<<ISC01)|(1<<ISC00)
#define KB_INT_BITNR        PD2

#define DDR_KB_INT          DDRD
#define PORT_KB_INT         PORTD
#define PIN_KB_INT          PIND
#define KB_INT_BIT          (1<<KB_INT_BITNR)
#define KB_CLK_GET()      (PIN_KB_INT  & KB_INT_BIT)

// Keyboard Clocking - Output
#define DDR_KB_CLK_OUT      DDRD
#define PORT_KB_CLK_OUT     PORTD
#define KB_CLK_OUT_BITNR    PD6
#define KB_CLK_OUT_BIT      (1<<KB_CLK_OUT_BITNR)


#define KB_DATA_SET(val)  do {\
                            if((val))\
                             PORT_KB_DATA_OUT &= ~KB_DATA_OUT_BIT;\
                            else\
                             PORT_KB_DATA_OUT |= KB_DATA_OUT_BIT;\
                          } while(0)
#define KB_CLK_SET(val)   do {\
                            if((val))\
                             PORT_KB_CLK_OUT  &= ~KB_CLK_OUT_BIT;\
                            else\
                             PORT_KB_CLK_OUT  |= KB_CLK_OUT_BIT;\
                          } while(0)

/* -------------------- Data definitions ----------------------- */
#define BUFF_SIZE     16
#define STATE_RECEIVE 0
#define STATE_SEND    1
static volatile unsigned char sr_state;          // 1 = send, 0 = receive

static struct
{
        unsigned is_up      : 1;
        unsigned num_lock   : 1;
        unsigned extend     : 1;     // E0
        unsigned shift_lock : 1;
        unsigned alt        : 1;
        unsigned alt_gr     : 1;
        unsigned left_ctrl  : 1;
        unsigned right_ctrl : 1;
} flags = { 0,1,0,0,0,0,0,0 };

static signed char  bitcount;
static unsigned int kb_buffer[BUFF_SIZE];
static unsigned int *inpt, *outpt;
static volatile unsigned char buffcnt;
static unsigned int  shift  = 0;
static unsigned char parity = 0;
static unsigned char data;                // Holds the received scan code


static unsigned char data2send;
static volatile unsigned char resend_flag = 0;
static volatile unsigned char kbAckFlag;


#define TX_ACK   0
#define TX_STOP  1
#define TX_PTY   2
#define TX_B7    3
#define TX_B6    4
#define TX_B5    5
#define TX_B4    6
#define TX_B3    7
#define TX_B2    8
#define TX_B1    9
#define TX_B0    10
#define TX_START 11

static unsigned char ledState;
#define SCRLOCK_LED  1
#define NUMLOCK_LED  2
#define CAPSLOCK_LED 4

/* --------------------- Code ---------------------- */
void checkResend(void);
void i_send_kbd(unsigned char send_data);


void kb_init(void)
{
    KB_INT_MASKREG    &= ~(1<<KB_CLOCKINT);   // Disable INT Interrupt // GIMSK
    sr_state = STATE_RECEIVE;
    bitcount = 11;

    inpt =  kb_buffer;              // Initialize buffer
    outpt = kb_buffer;
    buffcnt = 0;
    DDR_KB_DATA  &= ~KB_DATA_BIT;   // Data-Pin as input
    DDR_KB_INT   &= ~KB_INT_BIT;    // INT-Pin as input

    KB_DATA_SET(1);
    KB_CLK_SET(1);
    DDR_KB_DATA_OUT  |= KB_DATA_OUT_BIT;
    DDR_KB_CLK_OUT   |= KB_CLK_OUT_BIT;
    
    ANY_EDGE();                     // INT interrupt on any edge

    KB_INT_FLAGREG   = (1<<KB_CLOCKINT);      // Clear any pending Interrupt
    resend_flag = 0;
    KB_INT_MASKREG  |= (1<<KB_CLOCKINT);      // Enable INT Interrupt // GIMSK
//    i_send_kbd(0xFF);     // Reset KBD
}


/* signal handler for external interrupt int1 */
#if (__AVR_LIBC_VERSION__ < 10400UL)
SIGNAL(KB_INTERRUPT)
#else
ISR( KB_INTERRUPT )
#endif
{
  if(sr_state == STATE_RECEIVE)
  {
    if (!KB_CLK_GET())                       // Routine entered at falling edge
    {
        if(bitcount < 11 && bitcount > 2)    // Bit 3 to 10 is data. Parity bit,
        {                                    // start and stop bits are ignored.
            data >>= 1;
            if(KB_DATA_GET() != 0)
                data |= 0x80;
        }
    }
    else
    {                                        // Routine entered at rising edge
        if(--bitcount == 0)                  // All bits received
        {
            bitcount = 11;
            if (data == 0xAA)
               ;// Self Test completed sucessfully
            else if (data == 0xFE)
               resend_flag = 1;
            else if (data == 0xFA)
               kbAckFlag = 1;
            else

            // decode 
            if (!flags.is_up)               // Last data received was the up-key identifier
            {
                switch (data)
                {
                case 0xF0:                  // The up-key identifier
                    flags.is_up = 1;
                    break;
                case 0xE0:                  // Extended Character
                    shift |= EXTENDED;
                    flags.extend = 1;
                    break;
                case 0x58:                  // SHIFT-LOCK
                    shift ^= LOCK_SHIFT;
                    flags.shift_lock ^= 1;
                    // now, toggle the shift lock LED on the keyboard
                    ledState ^= CAPSLOCK_LED;
                    break;
                case 0x12:                  // Left SHIFT
                    shift |= LEFT_SHIFT;
                    // deactivate shift lock
                    shift &= ~LOCK_SHIFT;
                    flags.shift_lock = 0;
                    ledState &= ~CAPSLOCK_LED;
                    break;
                case 0x59:                  // Right SHIFT
                    shift |= RIGHT_SHIFT;
                    // deactivate shift lock
                    shift &= ~LOCK_SHIFT;
                    flags.shift_lock = 0;
                    ledState &= ~CAPSLOCK_LED;
                    break;

                case 0x14:                  // Ctrl
                    if(flags.extend)
                    {
                        shift |= RIGHT_CTRL;
                        flags.right_ctrl = 1;
                    }
                    else
                    {
                        shift |= LEFT_CTRL;
                        flags.left_ctrl = 1;
                    }
                    break;
                case 0x11:                  // Alt
                    if(flags.extend)
                    {
                        shift |= RIGHT_ALT;
                        flags.alt_gr = 1;
                    }
                    else
                    {
                        shift |= LEFT_ALT;
                        flags.alt = 1;
                    }
                    break;
#if 0
                case 0x??:                  // Windows-Key
                    shift |= WINDOWS_KEY;
                    break;
#endif
                case 0x77: // NUMLOCK
                    ledState ^= NUMLOCK_LED;
                    flags.num_lock ^= 1;
                    break;
                case 0x7E: // ScrollLock
                    ledState ^= SCRLOCK_LED;
                // fall through    break;
                default:
#if 0
                    if(flags.alt && flags.left_ctrl)    // ctrl-alt+Nummber = display-Mode-switch
                    {
			extern uint8_t mode;
                        switch(data)
                        {
                        case 0x45: mode = 0; break;   // <strg><alt>0
                        case 0x16: mode = 1; break;   // <strg><alt>1
                        case 0x1E: mode = 2; break;   // ...
                        case 0x26: mode = 3; break;
                        case 0x25: mode = 4; break;
                        case 0x2E: mode = 5; break;
                        case 0x36: mode = 6; break;
                        case 0x3D: mode = 7; break;
                        case 0x3E: mode = 8; break;
                        case 0x46: mode = 9; break;
                        default:
                            goto putInBuffer;
                        }
                    }
                    else
#endif
                    {
                      // putInBuffer:
                        if (buffcnt < BUFF_SIZE)                    // If buffer not full
                        {
                            *inpt = (shift | data);                 // Put scancode into buffer
                            inpt++;                                 // Increment pointer
                            buffcnt++;
                            if (inpt >= (kb_buffer + BUFF_SIZE))
                                inpt = kb_buffer;
                        }
                    }
                    shift &= ~(EXTENDED);
                    flags.extend = 0;
                    break;
                }
            }
            else        //
            {
                flags.is_up = 0;        // Two 0xF0 in a row not allowed
                switch (data)
                {
                case 0xE0:              // Extended Character
                    shift &= ~EXTENDED;
                    flags.extend = 0;
                    break;
                case 0xF0:              // The up-key identifier
                    flags.is_up = 1;
                    break;
                case 0x58:              // SHIFT-LOCK
                    break;              // ignore
                case 0x12:              // Left SHIFT
                    shift &= ~LEFT_SHIFT;
                    break;
                case 0x59:              // Right SHIFT
                    shift &= ~RIGHT_SHIFT;
                    break;

                case 0x14:              // Ctrl
                    if(flags.extend)
                    {
                        shift &= ~RIGHT_CTRL;
                        flags.right_ctrl = 0;
                        flags.extend = 0;
                    }
                    else
                    {
                        shift &= ~LEFT_CTRL;
                        flags.left_ctrl = 0;
                    }
                    break;
                case 0x11:              // Alt
                    if(flags.extend)
                    {
                        shift &= ~RIGHT_ALT;
                        flags.alt_gr = 0;
                        flags.extend = 0;
                    }
                    else
                    {
                        shift &= ~LEFT_ALT;
                        flags.alt = 0;
                    }
                    break;
#if 0
                case 0x??:              // Windows-Key
                    shift &= ~WINDOWS_KEY;
                    break;
#endif
                }
            }
        }
    }
  }
  else // sr_state == SEND
  {
    if (!KB_CLK_GET())                  // Routine entered at falling edge
    {
        bitcount--;                     // Next bit
        if(bitcount > TX_PTY)           // Data bits
        {
          KB_DATA_SET(data & 0x01);
          if(data & 1)
            parity++;
          data >>= 1;
        }
        else if(bitcount == TX_PTY)     // Parity bit
        {
          KB_DATA_SET(!(parity & 0x01));
        }
        else if(bitcount == TX_STOP)    // Stop bit
        {
            KB_DATA_SET(1);
        }
    }
    else
    {                                        // Routine entered at rising edge
        if(bitcount == TX_ACK)               // Rising edge after keyboards ACK
        {
            sr_state = STATE_RECEIVE;        // We are receiving again
            bitcount = 11;
            parity   = 0;
            kbAckFlag = 0;
        }
    }
  }
}


/* Send a two byte command to the keyboard */
void send_kbd(unsigned char send_data, unsigned char send_data1)
{
  // Wait while keyboard busy
  while((sr_state != STATE_RECEIVE) || (resend_flag) || (bitcount != 11))
    wdt_reset();

  // An ACK-byte is expected after the first byte 
  kbAckFlag = 0;
  i_send_kbd(send_data);
  while(!kbAckFlag)	// wait for the ACK-byte
  {
    checkResend();
    wdt_reset();
  }
  i_send_kbd(send_data1);
}

/* Initiate the actual send operation */
void i_send_kbd(unsigned char send_data)
{    
    while(resend_flag)	// ISR busy
       wdt_reset();

    KB_INT_MASKREG    &= ~(1<<KB_CLOCKINT);   // Disable keyboard interrupt

    // 1. Set data line high impedance -> high
    //KB_DATA_SET(1);

    // 2. set clock pin low
    KB_CLK_SET(0);

    // 3. wait at least 64 us
    data2send = send_data;   // Remember this byte in case of "resend" reply from kbd
    data      = send_data;   // Actual data used by the ISR
    sr_state  = STATE_SEND;
    bitcount  = TX_START;
    parity    = 0;
    flags.extend = 0;        // Forget any magled keyboard state,
                             //  the complete sequence will be repeated.
    _delay_us(64);

    // 4. pull data line low
    KB_DATA_SET(0);

    // 5. release clock
    KB_CLK_SET(1);

    // Enable interrupt on kbd-clock
    KB_INT_FLAGREG   = (1<<KB_CLOCKINT);      // Clear pending interrupt
    KB_INT_MASKREG  |= (1<<KB_CLOCKINT);      // Enable keyboard clocking interrupt
}

/*
 * Check if the keyboard replied by a RESEND byte.
 * Update any changed keyboard LEDs.
 */
void checkResend(void)
{
  static char shlock = 0;
  
  if(resend_flag)
  {
    resend_flag = 0;
    i_send_kbd(data2send);
//    uart_print(" Resend\r\n");
  }
  if(ledState != shlock)
  {
      shlock = ledState;
      send_kbd(0xED, shlock);
  }
}

/* Decode the keyboar data */
unsigned int kb_decode(unsigned int data)
{
    unsigned char sc, c;
    prog_char *z;

    z = (prog_char *)&scanCodes[0];
    sc = data & 0xFF;
    data &= 0xFF00;
    while((c = pgm_read_byte(z)) != 0)
    {
        if(c == sc)
            break;
        z += 4;
    }
    if(sc == c)
    {
        z++;
        if (data & RIGHT_ALT)                               // Alt Gr
        {
            // [3]
            return (data | pgm_read_byte(&z[2]));
        }
        if (data & (LEFT_CTRL|RIGHT_CTRL))                  // Ctrl
        {
            // [1]
            sc = pgm_read_byte(z);
            if(sc >= 'a' && sc <= 'z')
                return (data | (sc - 'a' + 1));
        }
        if (data & (LEFT_SHIFT|RIGHT_SHIFT|LOCK_SHIFT))     // Shifted
            z++;
        return (data | pgm_read_byte(z));                   // or Unshifted
    }
    return data;
}

unsigned int kb_peekcode(void)
{
    checkResend();
    if(buffcnt == 0)
        return 0;
    return (*outpt);                              // Get scancode
}

unsigned int kb_peekchar(void)
{
    checkResend();
    if(buffcnt == 0)
        return 0;
    return (kb_decode(*outpt));                  // Get character
}


unsigned int kb_getcode(void)
{
    unsigned int word;

    do
    {
        wdt_reset();
        checkResend();
    } while(buffcnt == 0);                         // Wait for data
    word = *outpt;                              // Get scancode
    outpt++;                                    // Increment pointer
    if (outpt >= (kb_buffer + BUFF_SIZE))
        outpt = kb_buffer;
    buffcnt--;                                  // Decrement buffer count
    return word;
}


unsigned int kb_getchar(void)
{
    checkResend();
    return kb_decode(kb_getcode());
}

unsigned char kbhit(void)
{
    checkResend();
    return buffcnt;
}

void kb_flush(void)
{
    unsigned char sreg = SREG;
    
    checkResend();
    cli();
    inpt    = kb_buffer;                      // Re-Initialize buffer
    outpt   = kb_buffer;
    buffcnt = 0;
    SREG = sreg;
    checkResend();
}

#if 1

int main(void)
{
   int k;

    kb_init();
    sei();
    uart_init(9600, 'N', 8, 1);
    uart_print("\r\nStarting to echo DATA from keyboard ...\r\n");
    ledState |= NUMLOCK_LED;
    while(1)
    {
        if(kbhit())
        {
           k = kb_getchar() & 0xFF;
           uart_send(k);        
#if 0
        if(k == '0')
        { send_kbd(0xED, 0); }
        else if (k == '1')
        { send_kbd(0xED, 1);  }
        else if (k == '2')
        { send_kbd(0xED, 2);  }
        else if (k == '3')
        { send_kbd(0xED, 3);  }
        else if (k == '4')
        { send_kbd(0xED, 4);  }
        else if (k == '5')
        { send_kbd(0xED, 5);  }
        else if (k == '6')
        { send_kbd(0xED, 6);  }
        else if (k == '7')
        { send_kbd(0xED, 7);  }
#endif   
        }  
    }
    return 0;
}

#endif

